==> gm-1.6.3_Linux.tar.gz: gm-1.6.3_Linux/drivers/linux/gm/gm_arch_vmops.c <==
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.	      				 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

#include "gm_internal.h"

#include <linux/mm.h>
#include <linux/file.h>

static void
gm_linux_detach_file (struct vm_area_struct *vma)
{
#if LINUX_XX >= 24
  gm_linux_vm_ops_t *ops = (gm_linux_vm_ops_t *) vma->vm_ops;
  struct gm_port_state * ps = ops->gm.ps;
  if (vma->vm_file && vma->vm_file == ps->arch.file)
    {
      vma->vm_file = 0;
      gm_assert (atomic_read (&ps->arch.file->f_count) > 1);
#if LINUX_XX >= 26
       list_del(&vma->shared);
#else
      if (vma->vm_next_share)
	vma->vm_next_share->vm_pprev_share = vma->vm_pprev_share;
      *vma->vm_pprev_share = vma->vm_next_share;
#endif
      fput(ps->arch.file);
    }
#endif
}

static void
gm_linux_vm_open (struct vm_area_struct *vma)
{
  gm_linux_vm_ops_t *ops = (gm_linux_vm_ops_t *) vma->vm_ops;
  gm_assert (gm_linux_vm_is_wrapped (vma));
  gm_assert (ops->gm.refcount > 0);

  if (vma->vm_mm != current->mm)
    {
      /* ok we are doing a fork, I do believe this is the only case
         where the equality does not hold in open */
      /* the kernel has duplicated the vma
	 and increased the reference of vm_file before calling us */
      gm_linux_detach_file (vma);
      vma->vm_ops = ops->gm.ori;
      gm_linux_fork (vma);
    }
  else
    {
      ops->gm.refcount += 1;
      if (ops->gm.ori && ops->gm.ori->open)
	ops->gm.ori->open (vma);
    }
}

static void
gm_linux_vm_detach (struct vm_area_struct *vma, gm_linux_vm_ops_t * ops)
{
  gm_assert ((gm_linux_vm_ops_t *) vma->vm_ops == ops);
  gm_assert (ops->gm.refcount > 0);

  ops->gm.refcount -= 1;
  gm_linux_detach_file (vma);
  vma->vm_ops = ops->gm.ori;
  if (!ops->gm.refcount)
    {
      int ref_count;
      struct gm_port_state * ps = ops->gm.ps;
      gm_arch_mutex_enter (&ps->sync);
      gm_assert (ps->arch.ops_count > 0);
      ps->arch.ops_count -= 1;
      ps->arch.ref_count -= 1;
      ref_count = ps->arch.ref_count;
      gm_assert (gm_hash_find (ps->arch.ops_hash, ops->gm.ori) ==
		 ops);
      gm_hash_remove (ps->arch.ops_hash, ops->gm.ori);
      gm_free (ops);
      gm_arch_mutex_exit (&ps->sync);
      if (!ref_count)
	gm_linux_port_close (ps);
    }
}

static void
gm_linux_vm_close (struct vm_area_struct *vma)
{
  gm_linux_vm_ops_t *ops = (gm_linux_vm_ops_t *) vma->vm_ops;
  gm_assert (gm_linux_vm_is_wrapped (vma));
  gm_assert (ops->gm.refcount > 0);

  gm_linux_vm_detach (vma, ops);

  if (vma->vm_ops && vma->vm_ops->close)
    vma->vm_ops->close (vma);
}

#if 0
/* could be used with 2.2.x to track adress space change, unfortunately impossible with 2.4.x */
static void
gm_linux_vm_unmap (struct vm_area_struct *vma, unsigned long start,
		   size_t size)
{
  gm_linux_vm_ops_t *ops = (gm_linux_vm_ops_t *) vma->vm_ops;
  gm_assert (ops->gm.refcount > 0);

  gm_assert (gm_linux_vm_is_wrapped (vma));
  /*  gm_unregister_port_memory(ops->gm.ps, start, size, 1); */
  if (ops->gm.ori && ops->gm.ori->unmap)
    ops->gm.ori->unmap (vma, start, size);
}
#endif


static gm_linux_vm_ops_t *
gm_linux_new_vm_ops (gm_port_state_t * ps, struct vm_operations_struct *ori)
{
  gm_linux_vm_ops_t *ops = gm_malloc (sizeof (*ops));
  gm_bzero (ops, sizeof (*ops));
  ops->gm.ori = ori;
  ops->gm.ps = ps;
  if (ori)
    ops->wrap = *ori;
  ops->wrap.open = gm_linux_vm_open;
  ops->wrap.close = gm_linux_vm_close;
#if 0
  ops->wrap.unmap = gm_linux_vm_unmap;
#endif
  return ops;
}


static gm_linux_vm_ops_t *
gm_linux_vm_get_ops (gm_port_state_t * ps, struct vm_area_struct *vma)
{
  gm_linux_vm_ops_t *ops = gm_hash_find (ps->arch.ops_hash, vma->vm_ops);
  gm_status_t status;
  if (!ops)
    {
      ops = gm_linux_new_vm_ops (ps, vma->vm_ops);
      status = gm_hash_insert (ps->arch.ops_hash, vma->vm_ops, ops);
      if (status != GM_SUCCESS)
	{
	  gm_free (ops);
	  return 0;
	}
      /* we are in ioctl mutex */
      ps->arch.ops_count += 1;
      ps->arch.ref_count += 1;
    }
  return ops;
}

int
gm_linux_vm_is_wrapped (struct vm_area_struct *vma)
{
  return vma->vm_ops && vma->vm_ops->open == gm_linux_vm_open;
}


gm_status_t
gm_linux_vm_wrap (gm_port_state_t * ps, struct vm_area_struct * vma)
{
  gm_linux_vm_ops_t *ops;
  gm_assert (!gm_linux_vm_is_wrapped (vma));
  if (ps->arch.mm != current->mm)
    return GM_INVALID_PARAMETER;
  ops = gm_linux_vm_get_ops (ps, vma);
  if (!ops)
    return GM_OUT_OF_MEMORY;
  vma->vm_ops = &ops->wrap;
  ops->gm.refcount += 1;
#if  LINUX_XX >= 24
  /* prevent Linux anonymous vma merge operations (see can_vma_merge
     in linux/mm.h)which unfortunately do not guarantee to call open
     and close method on the object, to do that we transform the
     anonymous vm_area into a file object */
  if (!vma->vm_file && ps->arch.file)
    {
      struct inode * inode = ps->arch.file->f_dentry->d_inode;
      struct address_space * mapping = inode->i_mapping;
      struct vm_area_struct **head;
      
      get_file (ps->arch.file);
      vma->vm_file = ps->arch.file;
#if LINUX_XX >= 26
       list_add(&vma->shared, &mapping->i_mmap);
#else
       head = &mapping->i_mmap;
      /* insert into the special inode mapping list */
      vma->vm_next_share = *head;
      if (*head)
	(*head)->vm_pprev_share = &vma->vm_next_share;
      *head = vma;
      vma->vm_pprev_share = head;
#endif
    }
#endif
  return GM_SUCCESS;
}

void
gm_linux_vm_unwrap_all (gm_port_state_t * ps)
{
  struct vm_area_struct *vma;
  if (ps->arch.ops_count == 0 || current->mm != ps->arch.mm)
    return;
  gm_always_assert (current->mm);
  gm_mmap_down_write (current->mm);
  vma = current->mm->mmap;
  while (vma)
    {
      if (gm_linux_vm_is_wrapped (vma))
	{
	  gm_linux_vm_ops_t * ops = (gm_linux_vm_ops_t *) vma->vm_ops;
	  if (ops->gm.ps == ps)
	    {
	      gm_linux_vm_detach (vma, (gm_linux_vm_ops_t *) vma->vm_ops);
	    }
	}
      vma = vma->vm_next;
    }
  vma = current->mm->mmap;
  gm_always_assert (ps->arch.ops_count == 0);
  while (vma)
    {
      gm_assert (gm_hash_find (ps->arch.ops_hash, vma->vm_ops) == NULL);
      vma = vma->vm_next;
    }
  gm_mmap_up_write (current->mm);
}

==> gm-2.0.8_Linux.tar.gz: gm-2.0.8_Linux/drivers/linux/gm/gm_arch_vmops.c <==
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.	      				 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

#include "gm_internal.h"
#include "gm_call_trace.h"

#include <linux/mm.h>
#include <linux/file.h>

static void
gm_linux_detach_file (struct vm_area_struct *vma)
{
#if LINUX_XX >= 24
  gm_linux_vm_ops_t *ops = (gm_linux_vm_ops_t *) vma->vm_ops;
  struct gm_port_state * ps = ops->gm.ps;
  if (vma->vm_file && vma->vm_file == ps->arch.file)
    {
      vma->vm_file = 0;
      gm_assert (atomic_read (&ps->arch.file->f_count) > 1);
#if LINUX_XX >= 26
       list_del(&vma->shared);
#else
      if (vma->vm_next_share)
	vma->vm_next_share->vm_pprev_share = vma->vm_pprev_share;
      *vma->vm_pprev_share = vma->vm_next_share;
#endif
      fput(ps->arch.file);
    }
#endif
}

static void
gm_linux_vm_open (struct vm_area_struct *vma)
{
  gm_linux_vm_ops_t *ops = (gm_linux_vm_ops_t *) vma->vm_ops;
  gm_assert (gm_linux_vm_is_wrapped (vma));
  gm_assert (ops->gm.refcount > 0);

  if (vma->vm_mm != current->mm)
    {
      /* ok we are doing a fork, I do believe this is the only case
         where the equality does not hold in open */
      /* the kernel has duplicated the vma
	 and increased the reference of vm_file before calling us */
      gm_linux_detach_file (vma);
      vma->vm_ops = ops->gm.ori;
      gm_linux_fork (vma);
    }
  else
    {
      ops->gm.refcount += 1;
      if (ops->gm.ori && ops->gm.ori->open)
	ops->gm.ori->open (vma);
    }
}

static void
gm_linux_vm_detach (struct vm_area_struct *vma, gm_linux_vm_ops_t * ops)
{
  GM_CALLED ();
  
  gm_assert ((gm_linux_vm_ops_t *) vma->vm_ops == ops);
  gm_assert (ops->gm.refcount > 0);

  ops->gm.refcount -= 1;
  gm_linux_detach_file (vma);
  vma->vm_ops = ops->gm.ori;
  if (!ops->gm.refcount)
    {
      int ref_count;
      struct gm_port_state * ps = ops->gm.ps;
      gm_arch_mutex_enter (&ps->sync);
      gm_assert (ps->arch.ops_count > 0);
      ps->arch.ops_count -= 1;
      ps->arch.ref_count -= 1;
      ref_count = ps->arch.ref_count;
      gm_assert (gm_hash_find (ps->arch.ops_hash, ops->gm.ori) ==
		 ops);
      gm_hash_remove (ps->arch.ops_hash, ops->gm.ori);
      gm_free (ops);
      gm_arch_mutex_exit (&ps->sync);
      if (!ref_count)
	gm_linux_port_close (ps);
    }

  GM_RETURN_NOTHING ();
}

static void
gm_linux_vm_close (struct vm_area_struct *vma)
{
  gm_linux_vm_ops_t *ops = (gm_linux_vm_ops_t *) vma->vm_ops;
  gm_assert (gm_linux_vm_is_wrapped (vma));
  gm_assert (ops->gm.refcount > 0);

  gm_linux_vm_detach (vma, ops);

  if (vma->vm_ops && vma->vm_ops->close)
    vma->vm_ops->close (vma);
}

#if 0
/* could be used with 2.2.x to track adress space change, unfortunately impossible with 2.4.x */
static void
gm_linux_vm_unmap (struct vm_area_struct *vma, unsigned long start,
		   size_t size)
{
  gm_linux_vm_ops_t *ops = (gm_linux_vm_ops_t *) vma->vm_ops;
  gm_assert (ops->gm.refcount > 0);

  gm_assert (gm_linux_vm_is_wrapped (vma));
  /*  gm_unregister_port_memory(ops->gm.ps, start, size, 1); */
  if (ops->gm.ori && ops->gm.ori->unmap)
    ops->gm.ori->unmap (vma, start, size);
}
#endif


static gm_linux_vm_ops_t *
gm_linux_new_vm_ops (gm_port_state_t * ps, struct vm_operations_struct *ori)
{
  gm_linux_vm_ops_t *ops;

  GM_CALLED ();
  
  ops = gm_malloc (sizeof (*ops));
  gm_bzero (ops, sizeof (*ops));
  ops->gm.ori = ori;
  ops->gm.ps = ps;
  if (ori)
    ops->wrap = *ori;
  ops->wrap.open = gm_linux_vm_open;
  ops->wrap.close = gm_linux_vm_close;
#if 0
  ops->wrap.unmap = gm_linux_vm_unmap;
#endif
  GM_RETURN (ops);
}


static gm_linux_vm_ops_t *
gm_linux_vm_get_ops (gm_port_state_t * ps, struct vm_area_struct *vma)
{
  gm_linux_vm_ops_t *ops;
  gm_status_t status;

  GM_CALLED ();

  ops = gm_hash_find (ps->arch.ops_hash, vma->vm_ops);
  if (!ops)
    {
      ops = gm_linux_new_vm_ops (ps, vma->vm_ops);
      status = gm_hash_insert (ps->arch.ops_hash, vma->vm_ops, ops);
      if (status != GM_SUCCESS)
	{
	  gm_free (ops);
	  GM_RETURN (0);
	}
      /* we are in ioctl mutex */
      ps->arch.ops_count += 1;
      ps->arch.ref_count += 1;
    }
  GM_RETURN (ops);
}

int
gm_linux_vm_is_wrapped (struct vm_area_struct *vma)
{
  return vma->vm_ops && vma->vm_ops->open == gm_linux_vm_open;
}


gm_status_t
gm_linux_vm_wrap (gm_port_state_t * ps, struct vm_area_struct * vma)
{
  gm_linux_vm_ops_t *ops;
  gm_assert (!gm_linux_vm_is_wrapped (vma));
  if (ps->arch.mm != current->mm)
    return GM_INVALID_PARAMETER;
  ops = gm_linux_vm_get_ops (ps, vma);
  if (!ops)
    return GM_OUT_OF_MEMORY;
  vma->vm_ops = &ops->wrap;
  ops->gm.refcount += 1;
#if  LINUX_XX >= 24
  /* prevent Linux anonymous vma merge operations (see can_vma_merge
     in linux/mm.h)which unfortunately do not guarantee to call open
     and close method on the object, to do that we transform the
     anonymous vm_area into a file object */
  if (!vma->vm_file && ps->arch.file)
    {
      struct inode * inode = ps->arch.file->f_dentry->d_inode;
      struct address_space * mapping = inode->i_mapping;
      struct vm_area_struct **head;
      
      get_file (ps->arch.file);
      vma->vm_file = ps->arch.file;
#if LINUX_XX >= 26
       list_add(&vma->shared, &mapping->i_mmap);
#else
       head = &mapping->i_mmap;
      /* insert into the special inode mapping list */
      vma->vm_next_share = *head;
      if (*head)
	(*head)->vm_pprev_share = &vma->vm_next_share;
      *head = vma;
      vma->vm_pprev_share = head;
#endif
    }
#endif
  return GM_SUCCESS;
}

void
gm_linux_vm_unwrap_all (gm_port_state_t * ps)
{
  struct vm_area_struct *vma;
  if (ps->arch.ops_count == 0 || current->mm != ps->arch.mm)
    return;
  gm_always_assert (current->mm);
  gm_mmap_down_write (current->mm);
  vma = current->mm->mmap;
  while (vma)
    {
      if (gm_linux_vm_is_wrapped (vma))
	{
	  gm_linux_vm_ops_t * ops = (gm_linux_vm_ops_t *) vma->vm_ops;
	  if (ops->gm.ps == ps)
	    {
	      gm_linux_vm_detach (vma, (gm_linux_vm_ops_t *) vma->vm_ops);
	    }
	}
      vma = vma->vm_next;
    }
  vma = current->mm->mmap;
  gm_always_assert (ps->arch.ops_count == 0);
  while (vma)
    {
      gm_assert (gm_hash_find (ps->arch.ops_hash, vma->vm_ops) == NULL);
      vma = vma->vm_next;
    }
  gm_mmap_up_write (current->mm);
}



/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
